/*
 * Copyright (c) 2008  Arjan van der Velde
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/*
 * Parts of this program are derived from the original FreeBSD iostat,
 * moused and camcontrol.
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <cam/scsi/scsi_message.h>
#include <camlib.h>

enum command_t {
	START = 1, STOP  = 0
};

static int scsistart(struct cam_device *device, int startstop,
					 int retry_count, int timeout);

int stopunit(const char * name, const int unit);
int stopunit_char(const char *dev); 

static int
scsistart(struct cam_device *device, int startstop,
		  int retry_count, int timeout)
{
	union ccb *ccb;
	int error = 0;
	
	ccb = cam_getccb(device);
	
	scsi_start_stop(&ccb->csio,
			/* retries */		retry_count,
			/* cbfcnp */		NULL,
			/* tag_action */	startstop ? MSG_SIMPLE_Q_TAG : MSG_ORDERED_Q_TAG,
			/* start/stop */	startstop,
			/* load_eject */	0,
			/* immediate */		0,
			/* sense_len */		SSD_FULL_SIZE,
			/* timeout */		timeout ? timeout : 120000);
	
	/* Disable freezing the device queue */
	ccb->ccb_h.flags |= CAM_DEV_QFRZDIS;
	
	/* Send command */
	if (cam_send_ccb(device, ccb) < 0) {
		perror("error sending start unit");
		cam_error_print(device, ccb, CAM_ESF_ALL, CAM_EPF_ALL, stderr);
		cam_freeccb(ccb);
		return(1);
	}
	
	/* Check success */
	if ((ccb->ccb_h.status & CAM_STATUS_MASK) == CAM_REQ_CMP)
	/* Success */
		if (startstop)
			fprintf(stdout, "Unit started successfully\n");
		else
			fprintf(stdout, "Unit stopped successfully\n");
		else {
			/* Error */
			error = 1;
			if (startstop)
				fprintf(stdout, "Error received from start unit command\n");
			else
				fprintf(stdout, "Error received from stop unit command\n");
			cam_error_print(device, ccb, CAM_ESF_ALL, CAM_EPF_ALL, stderr);
		}
	
	cam_freeccb(ccb);
	return(error);
}

int
stopunit(const char * name, const int unit) {
	int error = 0;
	int timeout = 0, retry_count = 1;
	struct cam_device *cam_dev = NULL;
	fprintf(stdout, "Stopping device %s%u\n", name, unit);
	if ((cam_dev = cam_open_spec_device(name, unit, O_RDWR, NULL)) == NULL) {
		error = 1;
		fprintf(stdout, "%s\n", cam_errbuf);
		return error;
	}
	error = scsistart(cam_dev, STOP, retry_count, timeout);
	cam_close_device(cam_dev);
	return error;
}

int
stopunit_char(const char *dev) {
	int unit = 0, error = 0;
	char name[30];
	if (cam_get_device(dev, name, sizeof name, &unit) == -1) {
		error = 1;
		fprintf(stdout, "%s\n", cam_errbuf);
		return error;
	}
	return stopunit(name, unit);
}

