/*
 * Copyright (c) 2008  Arjan van der Velde
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/*
 * Parts of this program are derived from the original FreeBSD iostat,
 * moused and camcontrol.
 */

#include <sys/resource.h>
#include <sys/cdefs.h>
#include <sys/param.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <devstat.h>
#include <unistd.h>
#include <time.h>
#include <syslog.h>
#include <errno.h>
#include <libutil.h>
#include <cam/scsi/scsi_message.h>
#include <camlib.h>
#include <err.h>
#include <signal.h>

#include "cam.h"
#include "log.h"
#include "spindown.h"


/*
 * Default values
 */
#define DEFAULT_MAX_IDLE_TIME 600 /* 10 minutes */
#define DEFAULT_INTERVAL 60 /* 1 minute */

/*
 * Update a statistic field in a dval structure.
 * If the new value for the field differs from
 * the old value:
 *
 *   Set the current value to the new value
 *   Set the timestamp field to the current time
 *   Change state to SPIN_UP if it is not SPIN_ERROR
 *
 * The latter is to prevent retrying over and over
 * again, once we've determined we cannot spin down
 * the device.
 *
 */
#define UPDATESTAT( j, var, field, dval, operations )	\
	if (dval[j].var != operations[field]) {		\
		dval[j].var = operations[field];	\
		if (dval[j].state != SPIN_ERROR)	\
			dval[j].state = SPIN_UP;	\
		time(&dval[j].ts);			\
	}

struct device_values {
	char* name;	/* device name */
	int unit;	/* unit number */
	time_t ts;	/* timestamp last measurement */
	unsigned int max_idle_time;
	state_t state;

	/* stats */
	uint64_t reads;
	uint64_t writes;
	uint64_t frees;
	uint64_t other;
};

struct { int n;
	 char *s;
} states[] = {
	{ SPIN_UP, "Spinning" },
	{ SPIN_DOWN, "Not spinning" },
	{ SPIN_ERROR, "Error" },
	{ SPIN_UNKNOWN, "Unknown" }
};

/*
 * For our convenience
 */
typedef struct device_values dval_t;
typedef struct statinfo statinfo_t;
typedef struct devinfo devinfo_t;
typedef struct devstat devstat_t;
typedef uint64_t stat_t;

statinfo_t sinfo;
volatile sig_atomic_t run, hupflag;
dval_t *dval;
int num_devices_specified;
int debug;
struct pidfh *pfh;

char *pidfile = "/var/run/spindown.pid";


/*
 * Local function declarations
 */
static void usage();
static void stopit();
static void hup();
static void dohup();

/*
 * Main loop
 */
int
main(int argc, char** argv) {

	int i = 0, c, j, retval;
	unsigned int max_idle_time = DEFAULT_MAX_IDLE_TIME;
	unsigned int interval = DEFAULT_MAX_IDLE_TIME;
	unsigned int num_devices = 0;
	char name[DEVSTAT_NAME_LEN + 1]; /* null terminated */
	struct sigaction si_sa;
	sigemptyset(&si_sa.sa_mask);
	si_sa.sa_flags = 0;
	hupflag = 0;
	time_t now;

	background = 0;
	debug = 0;
	
	/* Assign signal handler for SIGINT | SIGTERM */
	si_sa.sa_handler = stopit;
	if (sigaction(SIGINT | SIGTERM, &si_sa, 0) == -1) {
		logerrx(1, "sigaction SIGINT | SIGTERM");
	}
	
	/* Assign signal handler for SIGHUP */
	si_sa.sa_handler = hup;
	if (sigaction(SIGHUP, &si_sa, 0) == -1) {
		logerrx(1, "sigaction SIGHUP");
	}
	
	/*
	 * Make sure that the userland devstat version matches the kernel
	 * devstat version.  If not, exit and print a message informing
	 * the user of his mistake.
	 */
	if (devstat_checkversion(NULL) < 0)
		logerrx(1, "%s", devstat_errbuf);
	
	/* Process command line */
	num_devices_specified = 0;
	while ((c = getopt(argc, argv, "p:Dbt:i:d:")) != -1) {
		int optval;
		switch(c) {
			case 't': /* set max idle time for next device */
				optval = atoi(optarg);
				if (optval > 0)
					max_idle_time = (unsigned int) optval;
				else
					max_idle_time = DEFAULT_MAX_IDLE_TIME;
				break;
			case 'i': /* set check interval */
				optval = atoi(optarg);
				if (optval > 0)
					interval = optval;
				else
					interval = DEFAULT_INTERVAL;
				break;
			case 'd': /* add a device */
				num_devices_specified++;
				dval = (dval_t *) realloc(dval, sizeof(dval_t) * num_devices_specified);
				bzero(&dval[num_devices_specified - 1], sizeof(dval_t));
				retval = cam_get_device(optarg, name, sizeof name, 
						&(dval[num_devices_specified - 1].unit));
				if (retval != 0)
					logerrx(1, "cam_get_device() failed");
				dval[num_devices_specified - 1].name = strdup(name);
				dval[num_devices_specified - 1].max_idle_time = max_idle_time;
				dval[num_devices_specified - 1].state = SPIN_UNKNOWN;
				time(&dval[num_devices_specified - 1].ts);
				break;
			case 'b': /* daemonize */
				if (!debug)
					background = 1;
				break;
			case 'D': /* debugging mode */
				background = 0;
				debug = 1;
				break;
			case 'p': /* pidfile */
				pidfile = optarg;
				break;
			case '?':
				usage();
				exit(1);
			default:
				break;
		}
	}

#ifdef DEBUG
	for (i = 0; i < num_devices_specified; i++)
		printf("name: %s, unit %u\n", dval[i].name, dval[i].unit);
#endif

	/* exit is no devices */
	if (num_devices_specified == 0) {
		if (!background)
			usage();
		logerrx(1, "No devices specified!");
	}
	
	argc -= optind;
	argv += optind;

	/* init devinfo part of statinfo structure */
	sinfo.dinfo = (devinfo_t *) malloc(sizeof(devinfo_t));
	if (sinfo.dinfo == NULL)
		logerrx(1, "malloc failed");
	
	bzero(sinfo.dinfo, sizeof(struct devinfo));
	
	if (devstat_getdevs(NULL, &sinfo) == -1)
		logerrx(1, "%s", devstat_errbuf);
	
	num_devices = sinfo.dinfo->numdevs;
	run = 1;

	if (background) {
		pid_t mpid = 0;
		pfh = pidfile_open(pidfile, 0600, &mpid);
		if (pfh == NULL) {
			if (errno == EEXIST)
				logerrx(1, "spindown already running, pid: %d", mpid);
			logwarn("cannot open pid file");
		}
		if (daemon(0, 0)) {
			pidfile_remove(pfh);
			logerr(1, "failed to become a daemon");
		} else {
			pidfile_write(pfh);
		}
	}


	/* Main loop */	
	while (run) {

#ifdef DEBUG
		int olderrno = 0;
#endif

		/* record time */
		time(&now);

		/* refresh statistics */
		if (devstat_getdevs(NULL, &sinfo) == -1)
			logerrx(1, "%s", devstat_errbuf);
	
		/* update number, we may have caught a new device */
		num_devices = sinfo.dinfo->numdevs;

		/* match statistics to specified devices and
                   update stats / state if required */
		for (i = 0; i < num_devices; i++) {

 			devstat_t * device = &(sinfo.dinfo->devices[i]);
 			uint64_t * operations = device->operations;
		
#ifdef DEBUG
			olderrno = errno; errno = 0;
			logwarn("device name: %s", device->device_name);
			errno = olderrno;
#endif

			for (j = 0; j < num_devices_specified; j++) {

				if (strcmp(device->device_name, dval[j].name) == 0 &&
					device->unit_number == dval[j].unit) {

					/* Update stats, timestamps and states */
					UPDATESTAT(j, reads, DEVSTAT_READ, dval, operations);
					UPDATESTAT(j, writes, DEVSTAT_WRITE, dval, operations);
					UPDATESTAT(j, frees, DEVSTAT_FREE, dval, operations);
					UPDATESTAT(j, other, DEVSTAT_NO_DATA, dval, operations);

					/* If max idle time reached and still spinning... */
					if (now - dval[j].ts >= dval[j].max_idle_time && dval[j].state == SPIN_UP) {
						if (stopunit(dval[j].name, dval[j].unit) != 0) {
							logwarn("Error spinning down device: %s%u", dval[j].name, dval[j].unit);
							dval[j].state = SPIN_ERROR;
						} else {
							dval[j].state = SPIN_DOWN;
						}
					}

					if (debug) {
					printf("device: %s%u, max_idle_time: %u, rd: %lu, wr: %lu, frees: %lu, other: %lu, idle time: %ld, state: %s\n",
						   dval[j].name,
						   dval[j].unit,
						   dval[j].max_idle_time,
						   dval[j].reads,
						   dval[j].writes,
						   dval[j].frees,
						   dval[j].other,
						   now - dval[j].ts,
						   strstate(dval[j].state));
					}

				}

			}

		}

		sleep(interval);
		if (hupflag) dohup(); /* Handle SIGHUP signal */

	}
	
	/* Free names first! */
	for (i = 0; i < num_devices_specified; i++) {
		free(dval[i].name);
	}
	free(dval);
	free(sinfo.dinfo);
	return 0;

}

/*
 * Show usage
 */
static void
usage() {
	printf("\n");
	printf("Usage:\n");
	printf("\n");
	printf("spindown [ -D ] [ -b ] [ -p pidfile ] [ -d device ] [ -t seconds ]\n");
	printf("         [ -i seconds ]\n");
	printf("\n");
	printf("        -D runs spindown in debug mode, preventing it from daemonizing\n");
	printf("        -b makes spindown run in the background\n");
	printf("        -p pidfile\n");
	printf("        -d a device to be monitored (multiple devices may be specified)\n");
	printf("           The idle time at which the drive will spin down is the time\n");
	printf("           specified by the last -t parameter before this -d option. This\n");
	printf("           way idle times can be set independently for each device.\n");
	printf("        -t idle time at which a device will spin down (defaults to 600 seconds)\n");
	printf("           this option can be specified more than one. -d options specified\n");
	printf("           will use the last -t option encountered before the -d option itself.\n");
	printf("        -i check interval. This is the interval between device activity\n");
	printf("           checks (defaults to 60 seconds)\n");
	printf("\n");
}

/*
 * Signal handler, to stop the daemon
 */
static void
stopit() {
	run = 0;
}

/*
 * Signal handler for SIGHUP
 */
static void
hup() {
	hupflag = 1;
}

/*
 * Log current info on devices, and
 * reset error states if any
 */
static void
dohup() {
	int i;
	time_t now;
	time(&now);
	errno = 0;
	for (i = 0; i < num_devices_specified; i++) {
		logwarn("device: %s%u, max_idle_time: %u, rd: %lu, wr: %lu, frees: %lu, other: %lu, idle time: %ld, state: %s\n",
			dval[i].name,
			dval[i].unit,
			dval[i].max_idle_time,
			dval[i].reads,
			dval[i].writes,
			dval[i].frees,
			dval[i].other,
			now - dval[i].ts,
			strstate(dval[i].state));
		if (dval[i].state == SPIN_ERROR)
			dval[i].state = SPIN_UNKNOWN;
	}
	hupflag = 0;
}

char* strstate(state_t s) {
	int i = 0;
	for (i = 0; i < sizeof(states); i++)
		if (states[i].n == s) return states[i].s;
	return strstate(SPIN_UNKNOWN);
}


/* notes
 * Error state is to prevent the daemon from trying to spindown
 * a device if it did not succeed the first time. If the first
 * time fails, we asume the device is incapable of spinning down
 * (either it does not spin down at all or it does not spin down
 * using the cam API)
 * XXX: The error state should probably be renamed to something more
 * descriptive if it's used in this way.
 */


